import decimal
import logging
import sys
from typing import Any

from orjson import orjson

DEFAULT_SENSITIVE_FIELDS = ("password", "client_secret", "access_token")
SENSITIVE_VALUE_REPLACEMENT = "<sensitive>"


def default(obj):
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    raise TypeError


def log_json_data(data: dict[str, Any] | list[Any]):
    """
    Dumps dictionary or list data to json string. Values for sensitive field names are replaced with fixed value

    Args:
        data: Data to dump

    Returns:
        String representation of data if data is dictionary. If data is list returns list of string representation
    """
    if type(data) == list:
        first_element = data[0] if len(data) > 0 else None
        if first_element is None:
            return []
        if type(first_element) in (int, bool, float, str):
            return data
        return [
            _log_single_item(item) for item in data
        ]
    return _log_single_item(data)


def _log_single_item(data: dict[str, Any]) -> str:
    return orjson.dumps(
        {k: v if k not in DEFAULT_SENSITIVE_FIELDS else SENSITIVE_VALUE_REPLACEMENT for k, v in data.items()},
        option=orjson.OPT_INDENT_2, default=default).decode()


def get_stdout_logger(level: int = logging.INFO) -> logging.Logger:
    logger = logging.getLogger("lime_trader")
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(level=level)
    return logger



import datetime
from functools import partial
from logging import Logger
from typing import Callable, Iterator

from lime_trader.clients.account_feed_client import AccountFeedClient
from lime_trader.converters.cattr_converter import CAttrConverter
from lime_trader.handlers.account_feed_handler import AccountFeedHandler
from lime_trader.models.accounts import (AccountDetails, AccountPosition, AccountTrade, AccountTransaction,
                                         AccountTransactionsPage, AccountTradesPage, Route)
from lime_trader.api.authenticated_api_client import AuthenticatedApiClient
from lime_trader.models.page import PageRequest, Page
from lime_trader.constants.urls import (ACCOUNTS_GET_ACCOUNTS_BALANCES, ACCOUNTS_GET_ACCOUNT_POSITIONS,
                                        ACCOUNTS_GET_ACCOUNT_TRADES,
                                        ACCOUNTS_GET_ACCOUNT_TRANSACTIONS, ACCOUNTS_STREAMING_FEED, ACCOUNTS_GET_ROUTES)
from lime_trader.utils.pagination import iterate_pages


class AccountClient:
    """
    Contains methods related to account data
    """

    def __init__(self, api_client: AuthenticatedApiClient, logger: Logger):
        """
        Contains methods related to account data

        Args:
            api_client: API client that will be used to execute all requests
            logger: Logger used to submit client log messages
        """
        self._api_client = api_client
        self._logger = logger

    def get_balances(self) -> list[AccountDetails]:
        """
        Gets balance information for all user accounts as a list

        Returns:
            List of all user accounts with balance information
        """
        self._logger.debug("Getting account balances")
        return self._api_client.get(url=ACCOUNTS_GET_ACCOUNTS_BALANCES, path_params={},
                                    params={}, response_schema=list[AccountDetails])

    def get_positions(self, account_number: str, date: datetime.date | None = None,
                      strategy: bool | None = None) -> list[AccountPosition]:
        """
        Gets account positions for specified date, or today if date is not supplied

        Args:
            account_number: Account number
            date: Date for account positions. If set, returns incoming positions for specified date.
                  If not specified, the method returns current intraday positions
            strategy: Optional. If `true`, returns current intraday positions grouped by multi-leg strategies.
                      Doesn't work together with `date` parameter

        Returns:
            Account positions for specified date, or current intraday positions if date is not supplied
        """
        return self._api_client.get(url=ACCOUNTS_GET_ACCOUNT_POSITIONS,
                                    path_params={"account_number": account_number},
                                    params={"date": date, "strategy": strategy},
                                    response_schema=list[AccountPosition])

    def get_trades(self, account_number: str, date: datetime.date, page: PageRequest) -> Page[AccountTrade]:
        """
        Gets the trades history on the specified account, ordered by descending timestamp.

        Args:
            account_number: Account number
            date: Date for which to get trades
            page: Pagination parameters

        Returns:
            Page of account trades for the specified date
        """
        response = self._api_client.get(url=ACCOUNTS_GET_ACCOUNT_TRADES,
                                        path_params={"account_number": account_number, "date": date},
                                        params={"limit": page.size,
                                                "skip": page.get_offset()},
                                        response_schema=AccountTradesPage)
        return Page(data=response.trades, number=page.page, size=page.size, total_elements=response.count)

    def iterate_trades(self, account_number: str, date: datetime.date,
                       start_page: PageRequest) -> Iterator[Page[AccountTransaction]]:
        for page in iterate_pages(start_page=start_page,
                                  func=partial(self.get_trades, account_number=account_number,
                                               date=date)):
            yield page

    def get_routes(self, account_number: str) -> list[Route]:
        """
        Gets a list of all routes available for specified account

        Args:
            account_number: Account number
        Returns:
            List of all routes available for specified account
        """
        return self._api_client.get(url=ACCOUNTS_GET_ROUTES,
                                    path_params={"account_number": account_number}, params={},
                                    response_schema=list[Route])

    def get_transactions_journal(self, account_number: str, start_date: datetime.date | None,
                                 end_date: datetime.date | None,
                                 page: PageRequest) -> Page[AccountTransaction]:
        """
        Gets list of transactions and a counter of total transactions in the specified time period

        Args:
            account_number: Account number
            start_date: The period start date. Optional
            end_date: The period end date. Optional, current date by default
            page: Pagination parameters

        Returns:
            Page of account transactions for specified date period
        """
        response = self._api_client.get(url=ACCOUNTS_GET_ACCOUNT_TRANSACTIONS,
                                        path_params={"account_number": account_number},
                                        params={"limit": page.size,
                                                "skip": page.get_offset(),
                                                "start_date": start_date,
                                                "end_date": end_date}, response_schema=AccountTransactionsPage)

        return Page(data=response.transactions, number=page.page, size=page.size, total_elements=response.count)

    def iterate_transactions_journal(self, account_number: str, start_page: PageRequest,
                                     start_date: datetime.date | None = None,
                                     end_date: datetime.date | None = None,
                                     ) -> Iterator[Page[AccountTransaction]]:
        for page in iterate_pages(start_page=start_page,
                                  func=partial(self.get_transactions_journal, account_number=account_number,
                                               start_date=start_date, end_date=end_date)):
            yield page

    def _start_streaming_feed(self, on_message: Callable, on_error: Callable) -> AccountFeedClient:
        websocket_app = self._api_client.websocket_connection(url=ACCOUNTS_STREAMING_FEED,
                                                              path_params={}, on_error=on_error, on_message=on_message)
        client = AccountFeedClient(websocket_app=websocket_app, logger=self._logger)
        websocket_app.on_open = client.on_account_streaming_feed_open
        websocket_app.on_close = client.on_account_streaming_feed_close
        client.start()
        return client

    def stream_account_feed(self, callback_client: AccountFeedHandler) -> AccountFeedClient:
        return self._start_streaming_feed(
            on_error=callback_client.on_account_feed_client_internal_error,
            on_message=partial(callback_client.on_message, CAttrConverter()))



import json
import logging
import os

import urllib3.util
from dotenv import dotenv_values

from lime_trader.clients.account_client import AccountClient
from lime_trader.constants.config import DEFAULT_TIMEOUT
from lime_trader.converters.cattr_converter import CAttrConverter
from lime_trader.models.accounts import Credentials
from lime_trader.api.api_client import ApiClient
from lime_trader.api.authenticated_api_client import AuthenticatedApiClient
from lime_trader.clients.market_data_client import MarketDataClient
from lime_trader.clients.trading_client import TradingClient
from lime_trader.models.token_storage import ApiTokenStorage
from lime_trader.utils.logging_utils import get_stdout_logger


class LimeClient:
    """
    Main client that should be used for all communication with the API.

    Attributes:
        account: Account client
        market: Market client
        trading: Trading client
    """

    def __init__(self, base_url: str, credentials: Credentials, logger: logging.Logger | None = None,
                 auth_url: str | None = None, timeout: int = DEFAULT_TIMEOUT):
        """
        Args:
            base_url: Base url of the API, this will be prepended to each url
            credentials: Credentials object used for authenticating in the API
            logger: Logger to be used
            auth_url: URL of the authorization server. Only requests related to authentication/authorization
                     will use this url
            timeout: Request timeout
        """
        self._base_url = urllib3.util.parse_url(base_url)
        self._auth_url = urllib3.util.parse_url(auth_url) if auth_url is not None else self._base_url
        self._api_use_https = self._base_url.url.startswith("https://")
        self._api_use_wss = self._api_use_https
        self._auth_use_https = self._auth_url.url.startswith("https://")
        self._auth_use_wss = self._auth_use_https
        self._credentials = credentials
        self._logger = logger or get_stdout_logger()
        self._converter = CAttrConverter()
        self._api_client = ApiClient(base_url=self._base_url, converter=self._converter, use_https=self._api_use_https,
                                     use_wss=self._api_use_wss, logger=self._logger, timeout=timeout)
        self._auth_api_client = ApiClient(base_url=self._auth_url, converter=self._converter,
                                          use_https=self._auth_use_https,
                                          use_wss=self._auth_use_wss, logger=self._logger, timeout=timeout)
        self._token_storage = ApiTokenStorage(credentials=credentials, api_client=self._auth_api_client)
        self._authenticated_api_client = AuthenticatedApiClient(api_client=self._api_client,
                                                                token_storage=self._token_storage,
                                                                credentials=self._credentials, logger=self._logger)

        # instantiate clients
        self.account = AccountClient(api_client=self._authenticated_api_client, logger=self._logger)
        self.trading = TradingClient(api_client=self._authenticated_api_client, logger=self._logger)
        self.market = MarketDataClient(api_client=self._authenticated_api_client, logger=self._logger)

    @classmethod
    def from_json(cls, json_str: str, logger: logging.Logger | None = None) -> 'LimeClient':
        """
        Instantiates client from json string.

        Required JSON keys are:
            - username
            - password
            - client_id
            - client_secret
            - grant_type
            - base_url
            - auth_url

        Args:
            json_str: JSON string
            logger: Logger to use

        Returns:
            Configured LimeClient
        """
        data = json.loads(json_str)
        return cls.from_dict(data=data, logger=logger)

    @classmethod
    def from_file(cls, file_path: str, logger: logging.Logger | None = None) -> 'LimeClient':
        """
        Instantiates client from file. Contents of the file needs to be json

        Required JSON keys are:
            - username
            - password
            - client_id
            - client_secret
            - grant_type
            - base_url
            - auth_url

        Args:
            file_path: Path to the file
            logger: Logger to use

        Returns:
            Configured LimeClient
        """
        with open(file=file_path, mode="r") as f:
            content = f.read()
        return cls.from_json(json_str=content, logger=logger)

    @classmethod
    def from_dict(cls, data: dict[str, str], logger: logging.Logger | None = None) -> 'LimeClient':
        """
        Instantiates client from dictionary.
        Dictionary must have required keys:

        - username
        - password
        - client_id
        - client_secret
        - grant_type
        - base_url
        - auth_url

        Args:
            data: Dictionary with client data and credentials
            logger: Logger to use

        Returns:
            Configured LimeClient
        """
        credentials = Credentials(username=data["username"],
                                  password=data["password"],
                                  client_id=data["client_id"],
                                  client_secret=data["client_secret"],
                                  grant_type=data["grant_type"])
        client = cls(base_url=data["base_url"],
                     credentials=credentials,
                     auth_url=data["auth_url"],
                     logger=logger)
        return client

    @classmethod
    def from_env(cls, logger: logging.Logger | None = None) -> 'LimeClient':
        """
        Configures client by getting credentials and configuration from system environment.
        Required environment variables are:

        - LIME_SDK_USERNAME
        - LIME_SDK_PASSWORD
        - LIME_SDK_CLIENT_ID
        - LIME_SDK_CLIENT_SECRET
        - LIME_SDK_GRANT_TYPE
        - LIME_SDK_BASE_URL
        - LIME_SDK_AUTH_URL

        Args:
            logger: Logger to use

        Returns:
            Configured LimeClient
        """
        credentials = {
            "username": os.environ.get("LIME_SDK_USERNAME"),
            "password": os.environ.get("LIME_SDK_PASSWORD"),
            "client_id": os.environ.get("LIME_SDK_CLIENT_ID"),
            "client_secret": os.environ.get("LIME_SDK_CLIENT_SECRET"),
            "grant_type": os.environ.get("LIME_SDK_GRANT_TYPE"),
            "base_url": os.environ.get("LIME_SDK_BASE_URL"),
            "auth_url": os.environ.get("LIME_SDK_AUTH_URL"),
        }
        return cls.from_dict(credentials, logger)

    @classmethod
    def from_env_file(cls, file_path: str, logger: logging.Logger | None = None) -> 'LimeClient':
        """
        Configures client by getting credentials and configuration from key/value file, usually named `.env`.

        Required keys are:

        - LIME_SDK_USERNAME
        - LIME_SDK_PASSWORD
        - LIME_SDK_CLIENT_ID
        - LIME_SDK_CLIENT_SECRET
        - LIME_SDK_GRANT_TYPE
        - LIME_SDK_BASE_URL
        - LIME_SDK_AUTH_URL

        Args:
            file_path: File name from which to load config
            logger: Logger to use

        Returns:
            Configured LimeClient
        """
        config = dotenv_values(file_path)
        credentials = {
            "username": config.get("LIME_SDK_USERNAME"),
            "password": config.get("LIME_SDK_PASSWORD"),
            "client_id": config.get("LIME_SDK_CLIENT_ID"),
            "client_secret": config.get("LIME_SDK_CLIENT_SECRET"),
            "grant_type": config.get("LIME_SDK_GRANT_TYPE"),
            "base_url": config.get("LIME_SDK_BASE_URL"),
            "auth_url": config.get("LIME_SDK_AUTH_URL"),
        }
        return cls.from_dict(credentials, logger)



from typing import TypeVar, Type, Callable, Any

from urllib3.util import Url
from websocket import WebSocketApp

from lime_trader.api.authenticated_api_client import AuthenticatedApiClient

T = TypeVar("T")


class AsyncAuthenticatedApiClient(AuthenticatedApiClient):

    async def _get_authorization_header(self) -> dict[str, str]:
        return self._get_authorization_header_for_token(await self._token_storage.get_token())

    async def get(self, url: Url, path_params: dict[str, Any], params: Any, response_schema: Type[T]) -> T:
        return await self._api_client.get(url=url, path_params=path_params, params=params,
                                          headers=await self._get_authorization_header(),
                                          response_schema=response_schema)

    async def post_form(self, url: Url, path_params: dict[str, Any], data: Any, response_schema: Type[T]) -> T:
        return await self._api_client.post_form(url=url, path_params=path_params, data=data,
                                                headers=await self._get_authorization_header(),
                                                response_schema=response_schema)

    async def post(self, url: Url, path_params: dict[str, Any], json: Any, response_schema: Type[T]) -> T:
        return await self._api_client.post(url=url, path_params=path_params, json=json,
                                           headers=await self._get_authorization_header(),
                                           response_schema=response_schema)

    async def websocket_connection(self, url: Url, path_params: dict[str, Any], on_message: Callable,
                                   on_error: Callable,
                                   ) -> WebSocketApp:
        return self._api_client.websocket_connection(url=url, path_params=path_params,
                                                     on_message=on_message, on_error=on_error,
                                                     headers=await self._get_authorization_header(),
                                                     )



from logging import Logger
from typing import TypeVar, Type, Callable, Any

from urllib3.util import Url
from websocket import WebSocketApp

from lime_trader.models.accounts import Token, Credentials
from lime_trader.api.api_client import ApiClient
from lime_trader.models.token_storage import TokenStorage

T = TypeVar("T")


class AuthenticatedApiClient:
    """
    Used for communication with the API while being authenticated. It supports using different endpoint
    for authentication to support a case when authorization server is on different domain
    """

    def __init__(self, api_client: ApiClient, token_storage: TokenStorage, credentials: Credentials, logger: Logger):
        """
        Args:
            api_client: API client used for executing all requests
            token_storage: Token storage used for getting/setting token
            param credentials: Credentials used for API authentication
            logger: Used for logging information about executed requests
        """
        self._api_client = api_client
        self._token_storage = token_storage
        self._credentials = credentials
        self._logger = logger

    def _get_authorization_header_for_token(self, token: Token) -> dict[str, str]:
        """
        Constructs headers required for authorization with existing token.

        Args:
            token: Token information

        Returns:
            Headers required for authorization
        """
        return {"Authorization": f"Bearer {token.access_token}"}

    def _get_authorization_header(self) -> dict[str, str]:
        """
        Gets required authorization header. If token is already fetched, it uses it, if not, it fetches new token.

        Returns:
            Authorization headers
        """
        return self._get_authorization_header_for_token(self._token_storage.get_token())

    def get(self, url: Url, path_params: dict[str, Any], params: Any, response_schema: Type[T]) -> T:
        """
        Forwards GET request to underlying API client

        Args:
            url: Url path.
            path_params: URL path parameters
            params: GET parameters
            response_schema: Python type to which response data will be transformed

        Returns:
            Response converted to Python type passed as response_schema parameter
        """
        return self._api_client.get(url=url, path_params=path_params, params=params,
                                    headers=self._get_authorization_header(),
                                    response_schema=response_schema)

    def post_form(self, url: Url, path_params: dict[str, Any], data: Any, response_schema: Type[T]) -> T:
        """
        Forwards POST request with form data to underlying API client

        Args:
            url: Url path.
            path_params: URL path parameters
            data: Form data
            response_schema: Python type to which response data will be transformed

        Returns:
            Response converted to Python type passed as response_schema parameter
        """
        return self._api_client.post_form(url=url, path_params=path_params, data=data,
                                          headers=self._get_authorization_header(),
                                          response_schema=response_schema)

    def post(self, url: Url, path_params: dict[str, Any], json: Any, response_schema: Type[T]) -> T:
        """
        Forwards POST request with JSON data to underlying API client

        Args:
            url: Url path.
            path_params: URL path parameters
            json: JSON data
            response_schema: Python type to which response data will be transformed

        Returns:
            Response converted to Python type passed as response_schema parameter
        """
        return self._api_client.post(url=url, path_params=path_params, json=json,
                                     headers=self._get_authorization_header(),
                                     response_schema=response_schema)

    def websocket_connection(self, url: Url, path_params: dict[str, Any], on_message: Callable, on_error: Callable,
                             ) -> WebSocketApp:
        return self._api_client.websocket_connection(url=url, path_params=path_params,
                                                     on_message=on_message, on_error=on_error,
                                                     headers=self._get_authorization_header(),
                                                     )



import logging
from abc import abstractmethod
from logging import Logger
from typing import Any

import orjson
from websocket import WebSocketApp

from lime_trader.converters.abstract_converter import AbstractConverter
from lime_trader.models.accounts import (AccountPositions, AccountDetails, AccountTrade,
                                         AccountFeedType, AccountFeedError)
from lime_trader.models.trading import OrderDetails


class AccountFeedHandler:
    def __init__(self, logger: Logger | None = None, ):
        self._logger = logger if logger is not None else logging.getLogger()

    def on_account_feed_client_internal_error(self, websocket_app: WebSocketApp, error) -> None:
        self.on_client_error(error=error)

    def on_message(self, converter: AbstractConverter, websocket_app: WebSocketApp, message: str) -> None:
        self._logger.info("Received message on account feed")
        decoded_json = orjson.loads(message)
        message_type = decoded_json.get("t", None)
        if message_type is None:
            self._logger.error("Error decoding account feed message. Missing message type.")
            return
        decoded_type = AccountFeedType(message_type)
        decoded = decoded_json.get("data", None)
        if decoded is None:
            if decoded_type == AccountFeedType.ERROR:
                self.on_stream_error(converter.load_from_dict(decoded_json, AccountFeedError))
            else:
                self._logger.error("Error decoding account feed message. Missing message data.")
            return
        if decoded_type == AccountFeedType.TRADE:
            self.on_account_trades_execution(converter.load_from_dict(decoded, list[AccountTrade]))
        elif decoded_type == AccountFeedType.BALANCE:
            self.on_account_balances(converter.load_from_dict(decoded, list[AccountDetails]))
        elif decoded_type == AccountFeedType.ORDER:
            self.on_account_order_status_changes(converter.load_from_dict(decoded, list[OrderDetails]))
        elif decoded_type == AccountFeedType.POSITION:
            self.on_account_positions(converter.load_from_dict(decoded, list[AccountPositions]))
        else:
            raise ValueError(f"Unknown message type={message_type} for account feed!")

    def on_account_balances(self, accounts: list[AccountDetails]) -> None:
        for account in accounts:
            self.on_account_balance(account=account)

    def on_account_positions(self, account_positions: list[AccountPositions]) -> None:
        for position in account_positions:
            self.on_account_position(position=position)

    def on_account_order_status_changes(self, orders: list[OrderDetails]) -> None:
        for order in orders:
            self.on_account_order_status_change(order=order)

    def on_account_trades_execution(self, trades: list[AccountTrade]) -> None:
        for trade in trades:
            self.on_account_trade_execution(trade=trade)

    @abstractmethod
    def on_account_balance(self, account: AccountDetails) -> None:
        """
        The server sends a full list of balance for all subscribed accounts on each update of any field value.
        Since during market hours the prices keep changing the update is going to be sent on each throttling period

        Args:
            account: Account details that changed
        """
        self._logger.info(f"Balance: {account}")

    @abstractmethod
    def on_account_position(self, position: AccountPositions) -> None:
        """
        The server sends a full list of positions for all subscribed accounts on each update of any field value.
        Since during market hours the prices keep changing the update is going to be sent on each throttling period

        Args:
            position: Changed position
        """
        self._logger.info(f"Position: {position}")

    @abstractmethod
    def on_account_order_status_change(self, order: OrderDetails) -> None:
        """
        The server sends a notification when an order changes status.
        After initial connection, the server sends a list of all orders in current status on current trading day.

        Args:
            order: Updated order details
        """
        self._logger.info(f"Order: {order}")

    @abstractmethod
    def on_account_trade_execution(self, trade: AccountTrade) -> None:
        """
        The server sends a notification when a trade executes.
        After initial connection, the server sends a list of all trades on current trading day.

        Args:
            trade: Executed trade
        """
        self._logger.info(f"Trade: {trade}")

    @abstractmethod
    def on_stream_error(self, error: AccountFeedError) -> None:
        """
        Handles error returned as websocket message. This error indicates that message has been received and contains
        error message.

        Args:
            error: Decoded error
        """
        self._logger.info(f"Error: {error}")

    @abstractmethod
    def on_client_error(self, error: Any) -> None:
        """
        Handles client error, this error indicates error from client or error while decoding message.
        Message is received and cannot be decoded or there was error receiving message.

        Args:
            error: Error description
        """
        self._logger.info(f"Error: {error}")



from typing import Callable, Iterator, AsyncIterator

from lime_trader.models.page import Page, PageRequest


def iterate_pages(start_page: PageRequest, func: Callable) -> Iterator[Page]:
    """
    Creates iterator from callable and page. It assumes callable will have only one parameter, page

    Args:
        start_page: Start page
        func: Function that will be called with page parameter

    Returns:
        Iterator of pages
    """
    has_next_page = True
    current_page = PageRequest(page=start_page.page, size=start_page.size)
    while has_next_page:
        response = func(page=current_page)
        yield response
        has_next_page = response.has_next()
        if has_next_page:
            current_page.page += 1


async def iterate_pages_async(start_page: PageRequest, func: Callable) -> AsyncIterator[Page]:
    """
    Creates iterator from callable and page. It assumes callable will have only one parameter, page

    Args:
        start_page: Start page
        func: Function that will be called with page parameter

    Returns:
        Iterator of pages
    """
    has_next_page = True
    current_page = PageRequest(page=start_page.page, size=start_page.size)
    while has_next_page:
        response = await func(page=current_page)
        yield response
        has_next_page = response.has_next()
        if has_next_page:
            current_page.page += 1



from typing import TypeVar, Type, Any
import httpx
from urllib3.util import Url

from lime_trader.api.api_client import ApiClient
from lime_trader.exceptions.api_error import ApiError
from lime_trader.models.errors import Error
from lime_trader.utils.logging_utils import log_json_data

T = TypeVar("T")


class AsyncApiClient(ApiClient):

    async def get(self, url: Url, path_params: dict[str, Any], params: Any, headers: dict[str, str],
                  response_schema: Type[T]) -> T:
        request_url = self._get_url(url, path_params=path_params)
        dict_params = self._converter.dump_to_dict(params, remove_none=True)
        self._logger.debug(f"Executing GET request {request_url}, params={log_json_data(dict_params)}")
        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(url=request_url, params=dict_params, headers=headers, timeout=self._timeout)
            except httpx.TransportError as e:
                raise ApiError(
                    Error(status_code=None,
                          code="transport_error",
                          message=f"Transport error: {e.__class__.__name__}.{''.join(e.args)}")
                )
        return self._handle_response(response=response, response_schema=response_schema)

    async def post_form(self, url: Url, path_params: dict[str, Any], data: Any, headers: dict[str, str],
                        response_schema: Type[T]) -> T:
        request_url = self._get_url(url, path_params=path_params)
        dict_data = self._converter.dump_to_dict(data, remove_none=True)
        self._logger.debug(f"Executing POST request {request_url}, data={log_json_data(dict_data)}")
        async with httpx.AsyncClient() as client:
            try:
                response = await client.post(url=request_url, data=dict_data, headers=headers, timeout=self._timeout)
            except httpx.TransportError as e:
                raise ApiError(
                    Error(status_code=None,
                          code="transport_error",
                          message=f"Transport error: {e.__class__.__name__}.{''.join(e.args)}")
                )
        return self._handle_response(response=response, response_schema=response_schema)

    async def post(self, url: Url, path_params: dict[str, Any], json: Any, headers: dict[str, str],
                   response_schema: Type[T]) -> T:
        request_url = self._get_url(url, path_params=path_params)
        dict_data = self._prepare_json_body(json)
        self._logger.debug(f"Executing POST request {request_url}, json={log_json_data(dict_data)}")
        async with httpx.AsyncClient() as client:
            try:
                response = await client.post(url=request_url, json=dict_data, headers=headers, timeout=self._timeout)
            except httpx.TransportError as e:
                raise ApiError(
                    Error(status_code=None,
                          code="transport_error",
                          message=f"Transport error: {e.__class__.__name__}.{''.join(e.args)}")
                )
        return self._handle_response(response=response, response_schema=response_schema)



import datetime

DEFAULT_DATE_FORMAT = "%Y-%m-%d"


def date_to_str(date: datetime.date) -> str:
    """
    Converts date to string using specified format

    Args:
        date: Date to convert

    Returns:
        String representation of date in default format
    """
    return date.strftime(DEFAULT_DATE_FORMAT)


def str_to_date(date: str) -> datetime.date:
    """
    Converts string from default date format to date object

    Args:
        date: Date in string representation

    Returns:
        Date object
    """
    return datetime.datetime.strptime(date, DEFAULT_DATE_FORMAT).date()


def datetime_to_timestamp(date: datetime.datetime) -> int:
    """
    Converts datetime to unix timestamp

    Args:
        date: Datetime to convert

    Returns:
        Unix timestamp
    """
    return int(date.timestamp())


def timestamp_to_datetime(timestamp: int) -> datetime.datetime:
    """
    Converts unix timestamp to datetime object

    Args:
        timestamp: Unix timestamp

    Returns:
        Datetime object
    """
    if timestamp > 9999999999:
        # handle microseconds, this method will work properly till year 2899.
        timestamp = timestamp / 1000
    return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)



import datetime
import enum
from _decimal import Decimal
from dataclasses import dataclass


class Period(enum.Enum):
    MINUTE = "minute"
    MINUTE_5 = "minute_5"
    MINUTE_15 = "minute_15"
    MINUTE_30 = "minute_30"
    HOUR = "hour"
    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    QUARTER = "quarter"
    YEAR = "year"


class TradingSchedule(enum.Enum):
    PRE_MARKET = "pre_market"
    REGULAR_MARKET = "regular_market"
    AFTER_MARKET = "after_market"
    CLOSED = "closed"


class OptionType(enum.Enum):
    CALL = "call"
    PUT = "put"


@dataclass
class Quote:
    """
    Quote for the specified symbol.

    Attributes:
        symbol: The security symbol
        ask: Ask price
        ask_size: Ask size
        bid: Bid price
        bid_size: Bid size
        last: Last price
        last_size: Last trade size
        volume: Today total volume
        date: Last trade time
        high: Today's high price
        low: Today's low price
        open: Open price
        close: Yesterday's close price
        week52_high: 52-week high
        week52_low: 52-week low
        change: Today's price change
        change_pc: Today's percent price change
        open_interest: Open interest (options)
        implied_volatility: Implied volatility (options)
        theoretical_price: Theoretical price (options)
        delta: Delta value (options)
        gamma: Gamma value (options)
        theta: Theta value (options)
        vega: Vega value (options)
    """
    symbol: str
    ask: Decimal
    ask_size: Decimal
    bid: Decimal
    bid_size: Decimal
    last: Decimal
    last_size: Decimal
    volume: int
    date: datetime.datetime
    high: Decimal
    low: Decimal
    open: Decimal
    close: Decimal
    week52_high: Decimal
    week52_low: Decimal
    change: Decimal
    change_pc: Decimal
    open_interest: Decimal
    implied_volatility: Decimal
    theoretical_price: Decimal
    delta: Decimal
    gamma: Decimal
    theta: Decimal
    vega: Decimal


@dataclass
class QuoteHistory:
    timestamp: datetime.datetime
    period: Period
    open: Decimal
    high: Decimal
    low: Decimal
    close: Decimal
    volume: int


@dataclass
class CurrentSchedule:
    """Trading session info depending on current date and time

    Attributes:
        session: Current session info
    """

    session: TradingSchedule


@dataclass
class Security:
    """Represents security

    Attributes:
        symbol: Security symbol
        description: Description of security
    """

    symbol: str
    description: str


@dataclass
class SecuritiesPage:
    """Page of securities

    Attributes:
        trades: List of securities
        count: Total count of securities
    """

    trades: list[Security]
    count: int

    @property
    def securities(self) -> list[Security]:
        """
        Alias for returned list as API returns it as "trades". Should be used instead of "trades" attribute.

        Returns:
            List of securities
        """
        return self.trades


@dataclass
class Trade:
    timestamp: int
    quantity: int
    price: Decimal
    market: str


@dataclass
class TradesPage:
    """Represents one page of trades

    Attributes:
        trades: List of trades
        count: Total count of trades
    """

    trades: list[Trade]
    count: int



import datetime
from functools import partial
from logging import Logger
from typing import Callable, AsyncIterator

from lime_trader.clients.account_client import AccountClient
from lime_trader.clients.account_feed_client import AccountFeedClient
from lime_trader.models.accounts import (AccountDetails, AccountPosition, AccountTrade, AccountTransaction,
                                         AccountTransactionsPage, AccountTradesPage, Route)
from lime_trader.api.authenticated_api_client import AuthenticatedApiClient
from lime_trader.models.page import PageRequest, Page
from lime_trader.constants.urls import (ACCOUNTS_GET_ACCOUNTS_BALANCES, ACCOUNTS_GET_ACCOUNT_POSITIONS,
                                        ACCOUNTS_GET_ACCOUNT_TRADES,
                                        ACCOUNTS_GET_ACCOUNT_TRANSACTIONS, ACCOUNTS_STREAMING_FEED, ACCOUNTS_GET_ROUTES)
from lime_trader.utils.pagination import iterate_pages_async


class AsyncAccountClient(AccountClient):

    def __init__(self, api_client: AuthenticatedApiClient, logger: Logger):
        super().__init__(api_client=api_client, logger=logger)

    async def get_balances(self) -> list[AccountDetails]:
        return await self._api_client.get(url=ACCOUNTS_GET_ACCOUNTS_BALANCES,
                                          path_params={}, params={},
                                          response_schema=list[AccountDetails])

    async def get_positions(self, account_number: str, date: datetime.date | None = None,
                            strategy: bool | None = None) -> list[AccountPosition]:
        return await self._api_client.get(url=ACCOUNTS_GET_ACCOUNT_POSITIONS,
                                          path_params={"account_number": account_number},
                                          params={"date": date, "strategy": strategy},
                                          response_schema=list[AccountPosition])

    async def get_trades(self, account_number: str, date: datetime.date, page: PageRequest) -> Page[AccountTrade]:
        response = await self._api_client.get(url=ACCOUNTS_GET_ACCOUNT_TRADES,
                                              path_params={"account_number": account_number,
                                                           "date": date},
                                              params={"limit": str(page.size),
                                                      "skip": str(page.get_offset())},
                                              response_schema=AccountTradesPage)
        return Page(data=response.trades, number=page.page, size=page.size, total_elements=response.count)

    async def get_routes(self, account_number: str) -> list[Route]:
        return await self._api_client.get(url=ACCOUNTS_GET_ROUTES,
                                          path_params={"account_number": account_number}, params={},
                                          response_schema=list[Route])

    async def get_transactions_journal(self, account_number: str, start_date: datetime.date | None,
                                       end_date: datetime.date | None,
                                       page: PageRequest) -> Page[AccountTransaction]:
        response = await self._api_client.get(
            url=ACCOUNTS_GET_ACCOUNT_TRANSACTIONS,
            path_params={"account_number": account_number},
            params={"limit": page.size,
                    "skip": page.get_offset(),
                    "start_date": start_date,
                    "end_date": end_date}, response_schema=AccountTransactionsPage)

        return Page(data=response.transactions, number=page.page, size=page.size, total_elements=response.count)

    async def start_streaming_feed(self, on_message: Callable, on_error: Callable) -> AccountFeedClient:
        websocket_app = await self._api_client.websocket_connection(url=ACCOUNTS_STREAMING_FEED,
                                                                    path_params={}, on_error=on_error,
                                                                    on_message=on_message)
        client = AccountFeedClient(websocket_app=websocket_app, logger=self._logger)
        websocket_app.on_open = client.on_account_streaming_feed_open
        websocket_app.on_close = client.on_account_streaming_feed_close
        client.start()
        return client

    async def iterate_transactions_journal(self, account_number: str, start_page: PageRequest,
                                           start_date: datetime.date | None = None,
                                           end_date: datetime.date | None = None,
                                           ) -> AsyncIterator[Page[AccountTransaction]]:
        async for page in iterate_pages_async(start_page=start_page,
                                              func=partial(self.get_transactions_journal,
                                                           account_number=account_number,
                                                           start_date=start_date, end_date=end_date)):
            yield page



from dataclasses import dataclass


@dataclass(frozen=True)
class Error:
    """Error returned from API endpoints

    Attributes:
        status_code: Response status code
        code: Error code
        message: Error description
    """

    status_code: int | None
    code: str
    message: str



from lime_trader.models.errors import Error


class ApiError(Exception):
    """
    Exception thrown if error is returned from API
    """

    def __init__(self, error: Error):
        """
        Args:
            error: Error object containing details returned from API
        """
        self.error = error
        super().__init__()

    def __repr__(self) -> str:
        return repr(self.error)

    def __str__(self) -> str:
        return str(self.error)



from datetime import date, datetime, timedelta, timezone
import enum
from _decimal import Decimal
from functools import partial

from dataclasses import field, dataclass

from lime_trader.models.trading import TimeInForce, OrderType


class MarginType(enum.Enum):
    CASH = "cash"
    MARGIN_X1 = "marginx1"
    MARGIN_X2 = "marginx2"
    DAY_TRADER = "daytrader"


class RestrictionLevel(enum.Enum):
    NONE = "none"
    RESTRICTED = "restricted"
    DISABLED = "disabled"
    CLOSED = "closed"


class SecurityType(enum.Enum):
    COMMON_STOCK = "common_stock"
    PREFERRED_STOCK = "preferred_stock"
    OPTION = "option"
    STRATEGY = "strategy"


class TradeSide(enum.Enum):
    BUY = "buy"
    SELL = "sell"


class AccountFeedActionType(enum.Enum):
    SUBSCRIBE_BALANCE = "subscribeBalance"
    SUBSCRIBE_POSITIONS = "subscribePositions"
    SUBSCRIBE_ORDERS = "subscribeOrders"
    SUBSCRIBE_TRADES = "subscribeTrades"

    UNSUBSCRIBE_BALANCE = "unsubscribeBalance"
    UNSUBSCRIBE_POSITIONS = "unsubscribePositions"
    UNSUBSCRIBE_ORDERS = "unsubscribeOrders"
    UNSUBSCRIBE_TRADES = "unsubscribeTrades"


class AccountFeedType(enum.Enum):
    POSITION = "p"
    BALANCE = "b"
    ORDER = "o"
    TRADE = "t"
    ERROR = "e"


@dataclass(frozen=True)
class AccountDetails:
    """Represents account details and balances

    Attributes:
        account_number: Account number
        trade_platform: Trading platform this account is traded on
        margin_type: Margin type
        restriction: Restriction level effective on the account
        daytrades_count: Day trades counter
        account_value_total: Total account liquidation value
        cash: Account debit balance when negative, credit balance when positive
        day_trading_buying_power: Day trading buying power for marginable securities
        margin_buying_power: Buying power for marginable securities
        non_margin_buying_power: Buying power for non-marginable securities
        position_market_value: Sum of all positions current market values. The value is negative for short positions
        unsettled_cash: Unsettled cash for cash accounts
        cash_to_withdraw: Cash available to withdraw from the account
        restriction_reason: Optional description explaining why the account is restricted
    """

    account_number: str
    trade_platform: str
    margin_type: MarginType
    restriction: RestrictionLevel
    daytrades_count: int
    account_value_total: Decimal
    cash: Decimal
    day_trading_buying_power: Decimal
    margin_buying_power: Decimal
    non_margin_buying_power: Decimal

    position_market_value: Decimal

    unsettled_cash: Decimal
    cash_to_withdraw: Decimal
    restriction_reason: str | None = None


@dataclass(frozen=True)
class Leg:
    """Represents leg in the multi-leg strategy

    Attributes:
        symbol: The leg security symbol
        quantity: Signed number of shares or option contracts for the leg
        average_open_price: The leg cost basis
        current_price: The leg current price
        security_type: The leg asset type
    """
    symbol: str
    average_open_price: Decimal
    current_price: Decimal
    quantity: int
    security_type: SecurityType


@dataclass(frozen=True)
class AccountPosition:
    """Represents single account position

    Attributes:
        symbol: Security symbol
        quantity: Signed number of shares or option contracts. Negative for short positions
        average_open_price: Average historical cost basis
        current_price: Current price
        security_type: Asset type
        legs: Legs of the multi-leg strategy
    """

    symbol: str
    quantity: int
    average_open_price: Decimal
    current_price: Decimal
    security_type: SecurityType
    legs: list[Leg] | None = None


@dataclass(frozen=True)
class AccountPositions:
    """List of account positions

    Attributes:
        account: Account number
        positions: Account positions
    """

    account: str
    positions: list[AccountPosition]


@dataclass(frozen=True)
class AccountTrade:
    """Represents single account trade

    Attributes:
        symbol: Security symbol
        timestamp: Timestamp of the trade
        quantity: Number of shares or option contracts, negative for sells, positive for buys
        price: Trade price
        amount: Trade amount, which is the quantity multiplied by the lot size and price
        side: Trade side
    """

    symbol: str
    timestamp: datetime
    quantity: int
    price: Decimal
    amount: Decimal
    side: TradeSide


@dataclass(frozen=True)
class AccountTradesPage:
    """Page of account trades

    Attributes:
        trades: List of trades in page
        count: Total count of trades
    """
    trades: list[AccountTrade]
    count: int


@dataclass(frozen=True)
class Route:
    """Route available for account

    Attributes:
        exchange: Route name to use when placing order
        time_in_force: List of order duration instructions supported by a route
        order_type: List of order types supported by a route
    """
    exchange: str
    time_in_force: list[TimeInForce]
    order_type: list[OrderType]


@dataclass(frozen=True)
class TransactionCash:
    """Describes the cash side of the transaction

    Attributes:
        gross_amount: Dollar amount not including fees, can be positive or negative
        net_amount: Net dollar amount including fees charged for the transaction, can be positive or negative
    """

    gross_amount: Decimal
    net_amount: Decimal


@dataclass(frozen=True)
class TransactionAsset:
    """Describes the asset side of the transaction

    Attributes:
        symbol: Asset symbol
        symbol_description: Company name for stocks or human-readable name for options
        quantity: Transaction quantity, can be positive or negative
        price: Price for each unit
    """

    symbol: str
    symbol_description: str
    quantity: int
    price: Decimal


@dataclass(frozen=True)
class TransactionFee:
    """Represents transaction fee

    Attributes:
        name: Name of the fee.
        amount: Fee amount
    """

    name: str
    amount: Decimal


@dataclass(frozen=True)
class AccountTransaction:
    """Single account transaction

    Attributes:
        id: Internal transaction id, globally unique, and it is not necessarily sequentially incremented
        type: Transaction type
        description: Human-readable transaction description
        date: Date of transaction
        cash: Describes the cash side of the transaction
        fees: List of fees charged by the transaction
        asset: Structure describing the asset side of the transaction
    """

    id: str
    type: str
    description: str
    date: date
    cash: TransactionCash
    fees: list[TransactionFee]
    asset: TransactionAsset | None = None


@dataclass(frozen=True)
class AccountTransactionsPage:
    """Page of account transactions

    Attributes:
        transactions: List of transactions in page
        count: Total count of transactions
    """

    transactions: list[AccountTransaction]
    count: int


@dataclass(frozen=True)
class Token:
    """Represents access token

    Attributes:
        scope: The scopes this token grants access to
        token_type: Bearer means that access token should be put to the Authorization header of every web request
        access_token: Access token
        expires_in: Expiration lifetime in seconds
        date_created: Date when token was created. Used to determine expiry date
    """

    scope: str
    token_type: str
    access_token: str
    expires_in: int
    date_created: datetime = field(default_factory=partial(datetime.now, tz=timezone.utc))

    def expiry_date(self, tz: timezone = timezone.utc) -> datetime:
        """
        Returns expiry date in a specific timezone

        Args:
            tz: Timezone in which to return expiry date. Default is UTC

        Returns:
            Datetime when token expires
        """
        return self.date_created.astimezone(tz) + timedelta(seconds=self.expires_in)

    def expires_in_delta(self) -> timedelta:
        """
        Gets difference between expiry date and current date

        Returns:
            Difference between expiry date and current date
        """
        return self.expiry_date() - datetime.now(tz=timezone.utc)


@dataclass(frozen=True)
class Credentials:
    """Credentials used for API authentication

    Attributes:
        username: Username
        password: Password
        client_id: Client id issued to the service
        client_secret: Client secret issued to the service
        grant_type: OAuth authorization flow to use
    """
    username: str
    password: str
    client_id: str
    client_secret: str
    grant_type: str


@dataclass(frozen=True)
class AccountFeedAction:
    """Action that is sent to account feed websocket

    Attributes:
        action: Type of action
        account: Account number

    """
    action: AccountFeedActionType
    account: str


@dataclass(frozen=True)
class AccountFeedError:
    """Error during streaming account data

    Attributes:
        t: Type of account feed, always "e"
        code: Error code
        description: Error description
    """

    t: AccountFeedType
    code: str
    description: str



import datetime
from functools import partial
from logging import Logger
from typing import Iterator

from lime_trader.api.authenticated_api_client import AuthenticatedApiClient
from lime_trader.models.accounts import AccountTransaction
from lime_trader.models.market import (Quote, QuoteHistory, Period, Security, SecuritiesPage, Trade, TradesPage,
                                       CurrentSchedule)
from lime_trader.models.page import Page, PageRequest
from lime_trader.constants.urls import (MARKET_DATA_GET_CURRENT_QUOTE, MARKET_DATA_GET_QUOTES,
                                        MARKET_DATA_GET_TRADING_SCHEDULE,
                                        MARKET_DATA_LOOKUP_SECURITIES, MARKET_DATA_GET_TIME_AND_SALES,
                                        MARKET_DATA_GET_QUOTES_HISTORY)
from lime_trader.utils.pagination import iterate_pages


class MarketDataClient:
    """
    Contains methods related to market data
    """

    def __init__(self, api_client: AuthenticatedApiClient, logger: Logger):
        """
        Args:
            api_client: API client that will be used to execute all requests
            logger: Logger used to submit client log messages
        """
        self._api_client = api_client
        self._logger = logger

    def get_current_quote(self, symbol: str) -> Quote:
        """
        Retrieves current quote for the specified symbol

        Args:
            symbol: Security symbol

        Returns:
            Quote for the symbol
        """
        self._logger.info(f"Getting current quote for symbol {symbol}")
        return self._api_client.get(MARKET_DATA_GET_CURRENT_QUOTE, path_params={},
                                    params={"symbol": symbol}, response_schema=Quote)

    def get_current_quotes(self, symbols: list[str]) -> list[Quote]:
        """
        Retrieves current quotes for all specified symbols

        Args:
            symbols: List of security symbols

        Returns:
            List of quotes for specified symbols
        """
        self._logger.info(f"Getting current quotes for symbols {symbols}")
        return self._api_client.post(MARKET_DATA_GET_QUOTES, path_params={}, json=symbols, response_schema=list[Quote])

    def get_quotes_history(self, symbol: str, period: Period, from_date: datetime.datetime,
                           to_date: datetime.datetime) -> list[QuoteHistory]:
        """
        Returns candle structures aggregated by specified period

        Args:
            symbol: The security symbol, stocks in Nasdaq CMS convention. Options are not supported
            period: Aggregation period.
            from_date: Period start
            to_date: Period end

        Returns:
            Quote history between specified period
        """
        self._logger.info(f"Getting quotes history for symbol={symbol}, period={period}, from_date={from_date},"
                          f"to_date={to_date}")
        return self._api_client.get(MARKET_DATA_GET_QUOTES_HISTORY, path_params={},

                                    params={"symbol": symbol,
                                            "period": period.value,
                                            "from": from_date,
                                            "to": to_date
                                            }, response_schema=list[QuoteHistory])

    def get_trading_schedule(self) -> CurrentSchedule:
        """
        Returns trading session info depending on current date and time:

        Returns:
            Current trading schedule
        """
        self._logger.info(f"Getting current trading schedule")
        return self._api_client.get(MARKET_DATA_GET_TRADING_SCHEDULE, path_params={}, params={},
                                    response_schema=CurrentSchedule)

    def lookup_securities(self, query: str, page: PageRequest) -> Page[Security]:
        """
        Searches the securities reference by the specified criteria. The criteria can be a symbol,
        part of a symbol or part of a company name.

        Args:
            query: Search criteria
            page: Pagination parameters

        Returns:
            Page of security symbols which match criteria
        """
        self._logger.info(f"Looking up securities with query={query}, page={page}")
        response = self._api_client.get(MARKET_DATA_LOOKUP_SECURITIES, path_params={},
                                        params={"query": query,
                                                "limit": page.size,
                                                "skip": page.get_offset()},
                                        response_schema=SecuritiesPage)
        return Page(data=response.securities, number=page.page, size=page.size, total_elements=response.count)

    def time_and_sales(self, symbol: str, from_date: datetime.datetime,
                       to_date: datetime.datetime, page: PageRequest) -> Page[Trade]:
        """
        Retrieves the time and sales history, ordered by descending timestamp. Query parameters are:

        Args:
            symbol: The security symbol, stocks in Nasdaq CMS convention. Options are not supported
            from_date: Period start
            to_date: Period end
            page: Pagination parameters

        Returns:
            Page of historical trades
        """
        self._logger.info(f"Getting time and sales history for symbol={symbol}, from_date={from_date},"
                          f"to_date={to_date}, page={page}")
        response = self._api_client.get(MARKET_DATA_GET_TIME_AND_SALES, path_params={},
                                        params={"symbol": symbol,
                                                "from": from_date,
                                                "to": to_date,
                                                "limit": page.size,
                                                "skip": page.get_offset()},
                                        response_schema=TradesPage)
        return Page(data=response.trades, number=page.page, size=page.size, total_elements=response.count)

    def iterate_time_and_sales(self, symbol: str, from_date: datetime.datetime,
                               to_date: datetime.datetime, start_page: PageRequest,
                               ) -> Iterator[Page[AccountTransaction]]:
        """
        Returns iterator for the time and sales history, ordered by descending timestamp

        Args:
            symbol: The security symbol, stocks in Nasdaq CMS convention. Options are not supported
            from_date: Period start
            to_date: Period end
            start_page: Start page. Iterator starts from this page

        Returns:
            Iterator of pages of historical trades
        """
        for page in iterate_pages(start_page=start_page,
                                  func=partial(self.time_and_sales, symbol=symbol,
                                               from_date=from_date, to_date=to_date)):
            yield page



import datetime
from decimal import Decimal
from typing import Optional, List, Any, Type, TypeVar
import cattr

from lime_trader.converters.abstract_converter import AbstractConverter
from lime_trader.utils.date_utils import timestamp_to_datetime, datetime_to_timestamp, str_to_date, date_to_str

T = TypeVar("T")


class CAttrConverter(AbstractConverter):
    """
    Converter that utilizes cattr library to convert dataclasses from/to python objects
    """

    converter = cattr.Converter()
    dump_hidden_fields: Optional[List[str]] = None
    dump_specify_fields: Optional[List[str]] = None

    def __init__(self):
        # structure hook for load
        # unstructure hook for dump

        self.converter.register_structure_hook(Decimal, lambda val, _: Decimal(str(val)) if val is not None else None)
        self.converter.register_unstructure_hook(Decimal, lambda val: float(val) if val is not None else None)

        self.converter.register_structure_hook(datetime.datetime, lambda val, _: timestamp_to_datetime(
            int(val)) if val is not None else None)
        self.converter.register_unstructure_hook(datetime.datetime,
                                                 lambda val: datetime_to_timestamp(val) if val is not None else None)

        self.converter.register_structure_hook(datetime.date,
                                               lambda val, _: str_to_date(str(val)) if val is not None else None)
        self.converter.register_unstructure_hook(datetime.date,
                                                 lambda val: date_to_str(val) if val is not None else None)

    def dump_to_dict(self, t: Any, remove_none: bool = False) -> dict[str, Any]:
        if type(t) in (int, bool, str, float):
            return t
        d = self.converter.unstructure(t)
        if remove_none:
            return {k: v for k, v in d.items() if v is not None}
        return d

    def load_from_dict(self, data: dict[str, Any] | list[dict[str, Any]], t: Type[T]) -> T:
        return self.converter.structure(data, t)


converter = CAttrConverter()



import enum
from _decimal import Decimal
from dataclasses import dataclass


class OrderStatus(enum.Enum):
    NEW = "new"
    PENDING_NEW = "pending_new"
    PARTIALLY_FILLED = "partially_filled"
    FILLED = "filled"
    PENDING_CANCEL = "pending_cancel"
    CANCELED = "canceled"
    REPLACED = "replaced"
    REJECTED = "rejected"
    DONE_FOR_DAY = "done_for_day"
    SUSPENDED = "suspended"


class TimeInForce(enum.Enum):
    DAY = "day"
    EXT = "ext"


class OrderType(enum.Enum):
    MARKET = "market"
    LIMIT = "limit"


class OrderSide(enum.Enum):
    BUY = "buy"
    SELL = "sell"


@dataclass(frozen=True)
class Order:
    """
    Represents order

    Attributes:
        account_number: Account number this order belongs to.
        symbol: The security symbol, stocks in Nasdaq CMS convention, options in OCC.
        quantity: The positive decimal, number of shares or contracts.
        exchange: Optional, auto by default. The routing instructions for order execution.
                  The actual values are dynamic and depend on the account settings.
                  Some of the possible values are NASDAQ or ARCA
        price: Positive decimal if the order_type is limit or stop_limit
        time_in_force: Optional, day by default. Specifies how long the order remains in effect
        order_type: Optional, market by default
        side: Optional, buy by default. Available values are buy and sell.
              Buy opens long position, sell closes the position.
              The system will determine the proper side according to the current position,
              but you are still required to place two orders to revert the position from
              long to short and the other way around
    """

    account_number: str
    symbol: str
    quantity: Decimal
    exchange: str = "auto"
    price: Decimal | None = None

    time_in_force: TimeInForce = TimeInForce.DAY

    order_type: OrderType = OrderType.MARKET
    side: OrderSide = OrderSide.BUY

    def __post_init__(self):
        if (
                self.price is None
                or self.price < Decimal(0)
        ) and self.order_type in (OrderType.LIMIT, ):
            raise ValueError("Price must be positive decimal if the order_type is LIMIT")


@dataclass(frozen=True)
class OrderDetails:
    """
    Represents order details after order has been submitted

    Attributes:
        account_number: Account number this order belongs to
        client_id: Order id
        exchange: The routing instructions
        quantity: Number of shares or contracts requested by the order
        executed_quantity: Number of shares or contracts executed by this time
        order_status: Order status
        price: Limit price if applicable
        stop_price: Stop price if applicable
        time_in_force: Order duration instructions
        order_type: Type of the order
        order_side: Side of the order
        symbol: Security symbol
    """

    account_number: str
    client_id: str
    exchange: str
    quantity: Decimal
    executed_quantity: Decimal
    order_status: OrderStatus
    price: Decimal
    stop_price: Decimal
    time_in_force: TimeInForce
    order_type: OrderType
    order_side: OrderSide
    symbol: str

    @property
    def id(self) -> str:
        return self.client_id

    @property
    def order_id(self) -> str:
        return self.order_id


@dataclass(frozen=True)
class PlaceOrderResponse:
    """
    Order place response

    Attributes:
        success: Indicator if order place was success
        data: ID that is assigned to the order

    """

    success: bool
    data: str

    @property
    def order_id(self) -> str:
        return self.data


@dataclass(frozen=True)
class ValidateOrderResponse:
    """
    Order validation response

    Attributes:
        is_valid: Status of validation
        validation_message: Optional reject reason
    """

    is_valid: bool
    validation_message: str | None = None


@dataclass(frozen=True)
class CancelOrderResponse:
    """
    Order cancellation response

    Attributes:
        success: Indicator if cancel was success
        data: Cancellation request id
    """

    success: bool
    data: str


@dataclass(frozen=True)
class OrderFee:
    """
    Order fee

    Attributes:
        amount: Dollar amount
        type: Charge description
    """

    amount: Decimal
    type: str



from _decimal import Decimal
from logging import Logger

from lime_trader.api.authenticated_api_client import AuthenticatedApiClient

from lime_trader.clients.trading_client import TradingClient
from lime_trader.models.accounts import TradeSide
from lime_trader.models.trading import (Order, PlaceOrderResponse, ValidateOrderResponse, CancelOrderResponse,
                                       OrderFee, OrderDetails)
from lime_trader.constants.urls import (TRADING_GET_ACTIVE_ORDERS, TRADING_PLACE_ORDER, TRADING_VALIDATE_ORDER,
                                        TRADING_GET_ORDER_DETAILS,
                                        TRADING_CANCEL_ORDER, TRADING_ESTIMATE_FEE_CHARGES)


class AsyncTradingClient(TradingClient):

    def __init__(self, api_client: AuthenticatedApiClient, logger: Logger):
        super().__init__(api_client=api_client, logger=logger)

    async def place_order(self, order: Order) -> PlaceOrderResponse:
        return await self._api_client.post(url=TRADING_PLACE_ORDER,
                                           path_params={},
                                           json=order,
                                           response_schema=PlaceOrderResponse)

    async def validate_order(self, order: Order) -> ValidateOrderResponse:
        return await self._api_client.post(url=TRADING_VALIDATE_ORDER,
                                           path_params={},
                                           json=order,
                                           response_schema=ValidateOrderResponse)

    async def get_order_details(self, order_id: str) -> OrderDetails:
        return await self._api_client.get(url=TRADING_GET_ORDER_DETAILS,
                                          path_params={"id": order_id}, params={},
                                          response_schema=OrderDetails)

    async def cancel_order(self, order_id: str, message: str | None) -> CancelOrderResponse:
        return await self._api_client.post(url=TRADING_CANCEL_ORDER,
                                           path_params={"id": order_id},
                                           json={"message": message},
                                           response_schema=CancelOrderResponse)

    async def get_active_orders(self, account_number: str) -> list[OrderDetails]:
        return await self._api_client.get(url=TRADING_GET_ACTIVE_ORDERS,
                                          path_params={"account_number": account_number},
                                          params={},
                                          response_schema=list[OrderDetails])

    async def estimate_fee_charges(self, account_number: str, symbol: str, quantity: Decimal, side: TradeSide,
                                   price: Decimal) -> list[OrderFee]:
        return await self._api_client.post(url=TRADING_ESTIMATE_FEE_CHARGES, path_params={}, json={
            "account_number": account_number,
            "symbol": symbol,
            "quantity": quantity,
            "side": side,
            "price": price
        }, response_schema=list[OrderFee])



import logging
from json import JSONDecodeError
from logging import Logger
from typing import TypeVar, Type, Any, Callable
import httpx
import websocket
from httpx import Response, HTTPStatusError
from urllib3.util import Url

from lime_trader.converters.abstract_converter import AbstractConverter
from lime_trader.models.errors import Error
from lime_trader.exceptions.api_error import ApiError
from lime_trader.utils.logging_utils import log_json_data

T = TypeVar("T")


class ApiClient:
    """
    Used for communication with the API
    """

    def __init__(self, base_url: Url, converter: AbstractConverter, use_https: bool, use_wss: bool,
                 timeout: int,
                 logger: Logger):
        """
        Args:
            base_url: Base url that will be appended to all urls
            converter: Converter to use for transforming response data to Python objects
            use_https: Indicator if https protocol should be used, http is used if False
            use_wss: Indicator if wss protocol should be used, ws is used if False
            timeout: Request timeout
            logger: Logger to use for logging messages
        """
        self._base_url = base_url
        self._converter = converter
        self._use_https = use_https
        self._use_wss = use_wss
        self._timeout = timeout
        self._logger = logger

    def _map_response(self, data: dict[str, Any] | list[str, Any], t: Type[T]) -> T:
        """
        Does conversion between API response (json converted to dictionary) and Python type

        Args:
            data: Response data that needs to be converted
            t: Type to convert to

        Returns:
            Data converted to Python type passed as t parameter
        """
        return self._converter.load_from_dict(data, t)

    def _check_for_errors(self, response: Response) -> dict[str, Any]:
        """
        Checks for errors in a response. If there is, raises exception with returned error

        Args:
            response: Response to check for errors

        Returns:
            True if there is no error

        Raises:
            ApiError if response does not have success http status code
        """
        try:
            data = response.json()
        except JSONDecodeError as e:
            raise ApiError(
                Error(status_code=response.status_code,
                      code="json_decode_error",
                      message=e.msg)
            )
        try:
            response.raise_for_status()
        except HTTPStatusError as e:
            raise ApiError(
                Error(status_code=response.status_code,
                      code=data.get("code", ""),
                      message=data.get("message", ""))
            )

        return data

    def _handle_response(self, response: Response, response_schema: Type[T]) -> T:
        """
        Handles response by checking for errors and transforming data to appropriate Python type

        Args:
            response: Response returned from API
            response_schema: Python type to which response data will be transformed

        Returns:
            Pythonic type
        """
        response_data = self._check_for_errors(response)
        self._logger.debug(f"Response: {log_json_data(response_data)}")
        return self._map_response(response_data, response_schema)

    def get(self, url: Url, path_params: dict[str, Any], params: Any, headers: dict[str, str],
            response_schema: Type[T]) -> T:
        """
        Executes GET request with specified parameters

        Args:
            url: Url path.
            path_params: URL path parameters
            params: GET parameters
            headers: Request headers
            response_schema: Python type to which response data will be transformed

        Returns:
            Response converted to Python type passed as response_schema parameter
        """
        request_url = self._get_url(url, path_params=path_params)
        dict_params = self._converter.dump_to_dict(params, remove_none=True)
        self._logger.debug(f"Executing GET request {request_url}, params={log_json_data(dict_params)}")
        try:
            response = httpx.get(url=request_url, params=dict_params, headers=headers, timeout=self._timeout)
        except httpx.TransportError as e:
            raise ApiError(
                Error(status_code=None,
                      code="transport_error",
                      message=f"Transport error: {e.__class__.__name__}.{''.join(e.args)}")
            )
        return self._handle_response(response=response, response_schema=response_schema)

    def _get_url(self, path: Url, path_params: dict[str, Any]) -> str:
        """
        Gets full url by prefixing base url to path

        Args:
            path: Path to append to base url.
            path_params: URL path parameters

        Returns:
            Full url
        """
        if path.url.startswith("http://"):
            scheme = "https" if self._use_https else "http"
        elif path.url.startswith("ws://"):
            scheme = "wss" if self._use_wss else "ws"
        else:
            raise Exception("Invalid urls specified!")
        url = Url(scheme=scheme, path=(self._base_url.path or '') + path.path, host=self._base_url.hostname).url
        return url.format(**self._converter.dump_to_dict(path_params))

    def post_form(self, url: Url, path_params: dict[str, Any], data: Any, headers: dict[str, str],
                  response_schema: Type[T]) -> T:
        """
        Executes POST request with form data

        Args:
            url: Url path.
            path_params: URL path parameters
            data: Form data
            headers: Request headers
            response_schema: Python type to which response data will be transformed

        Returns:
            Response converted to Python type passed as response_schema parameter
        """
        request_url = self._get_url(url, path_params=path_params)
        dict_data = self._converter.dump_to_dict(data, remove_none=True)
        self._logger.debug(f"Executing POST request {request_url}, data={log_json_data(dict_data)}")
        try:
            response = httpx.post(url=request_url, data=dict_data, headers=headers, timeout=self._timeout)
        except httpx.TransportError as e:
            raise ApiError(
                Error(status_code=None,
                      code="transport_error",
                      message=f"Transport error: {e.__class__.__name__}.{''.join(e.args)}")
            )
        result = self._handle_response(response=response, response_schema=response_schema)
        return result

    def _prepare_json_body(self, json: Any) -> dict[str, Any] | list[dict[str, Any]]:
        if type(json) == list:
            return [self._converter.dump_to_dict(item, remove_none=True) for item in json]
        return self._converter.dump_to_dict(json, remove_none=True)

    def post(self, url: Url, path_params: dict[str, Any], json: Any, headers: dict[str, str],
             response_schema: Type[T]) -> T:
        """
        Executes POST request with JSON data

        Args:
            url: Url path.
            path_params: URL path parameters
            json: JSON data
            headers: Request headers
            response_schema: Python type to which response data will be transformed

        Returns:
            Response converted to Python type passed as response_schema parameter
        """
        request_url = self._get_url(url, path_params=path_params)
        dict_data = self._prepare_json_body(json)
        self._logger.debug(f"Executing POST request {request_url}, json={log_json_data(dict_data)}")
        try:
            response = httpx.post(url=request_url, json=dict_data, headers=headers, timeout=self._timeout)
        except httpx.TransportError as e:
            raise ApiError(
                Error(status_code=None,
                      code="transport_error",
                      message=f"Transport error: {e.__class__.__name__}.{''.join(e.args)}")
            )
        return self._handle_response(response=response, response_schema=response_schema)

    def websocket_connection(self, url: Url, path_params: dict[str, Any], on_message: Callable,
                             on_error: Callable,
                             headers: dict[str, str]) -> websocket.WebSocketApp:
        request_url = self._get_url(url, path_params=path_params)
        self._logger.debug(f"Opening websocket connection {request_url}")
        if self._logger.level <= logging.DEBUG:
            websocket.enableTrace(True)
        ws = websocket.WebSocketApp(url=request_url,
                                    on_message=on_message, on_error=on_error,
                                    header=headers)

        return ws



from datetime import timedelta

from lime_trader.api.api_client import ApiClient
from lime_trader.async_trader.api.async_api_client import AsyncApiClient
from lime_trader.constants.urls import AUTHENTICATION_GET_TOKEN
from lime_trader.models.accounts import Token, Credentials


class TokenStorage:

    def __init__(self):
        self._token = None

    def get_token(self) -> Token:
        return self._token

    def set_token(self, token: Token) -> None:
        self._token = token


class ApiTokenStorage(TokenStorage):

    def __init__(self, credentials: Credentials, api_client: ApiClient):
        super().__init__()
        self._credentials = credentials
        self._api_client = api_client

    def _get_token(self) -> Token:
        """
        Gets new token by executing API request

        Returns:
            Access token
        """
        return self._api_client.post_form(AUTHENTICATION_GET_TOKEN, path_params={}, data=self._credentials,
                                          response_schema=Token,
                                          headers={})

    def _validate_token(self, token: Token | None) -> bool:
        """
        Validates token by checking if it is fetched already or expires soon

        Args:
            token: Token to validate

        Returns:
            True if token is valid and can be used for authorization, False otherwise
        """
        if token is None or token.expires_in_delta() < timedelta(minutes=2):
            return False
        return True

    def get_token(self) -> Token:
        if not self._validate_token(token=self._token):
            self._token = self._get_token()
        return self._token

    def set_token(self, token: Token) -> None:
        self._token = token


class AsyncApiTokenStorage(ApiTokenStorage):

    def __init__(self, credentials: Credentials, api_client: AsyncApiClient):
        super().__init__(credentials=credentials, api_client=api_client)

    async def _get_token(self) -> Token:
        """
        Gets new token by executing API request

        Returns:
            Access token
        """
        return await self._api_client.post_form(AUTHENTICATION_GET_TOKEN, path_params={}, data=self._credentials,
                                                response_schema=Token,
                                                headers={})

    def _validate_token(self, token: Token | None) -> bool:
        """
        Validates token by checking if it is fetched already or expires soon

        Args:
            token: Token to validate

        Returns:
            True if token is valid and can be used for authorization, False otherwise
        """
        if token is None or token.expires_in_delta() < timedelta(minutes=2):
            return False
        return True

    async def get_token(self) -> Token:
        if not self._validate_token(token=self._token):
            self._token = await self._get_token()
        return self._token

    def set_token(self, token: Token) -> None:
        self._token = token



from dataclasses import dataclass
import math
from typing import TypeVar, Generic

# Generic page type
T = TypeVar("T")


@dataclass
class Page(Generic[T]):
    """
    Represents one page of items

    Attributes:
        data: Items in a page.
        number: Page number
        size: Page size
        total_elements: Number of existing elements. It does not represent number of elements in a page but
                        number of elements in total
    """

    data: list[T]
    number: int
    size: int

    total_elements: int

    def get_total_pages(self) -> int:
        """
        Get number of total pages based on total elements and page size

        Returns:
            Number of existing pages
        """
        if self.size == 0:
            return 1
        return int(math.ceil(self.total_elements / self.size))

    def has_next(self) -> bool:
        """
        Checks if there is a page after current one

        Returns:
            True if next page exists, False otherwise
        """
        return self.number < self.get_total_pages()

    def has_previous(self) -> bool:
        """
        Checks if there is a page before current one. Will be False just for the first page

        Returns:
            True if previous page exists, False otherwise
        """
        return 1 < self.number <= self.get_total_pages()

    def next_page_number(self) -> int:
        """
        Gets next page number

        Returns:
            Next page number
        """
        return self.number + 1

    def previous_page_number(self) -> int:
        """
        Gets previous page number

        Returns:
            Previous page number
        """
        return self.number - 1

    def is_last(self) -> bool:
        """
        Checks if current page is the last one

        Returns:
            True if current page is the last one, False otherwise
        """
        return not self.has_next()


@dataclass
class PageRequest:
    """Page specification

    Attributes:
        page: Page number
        size: Page size
    """

    page: int = 1
    size: int = 20

    def get_offset(self) -> int:
        """
        Gets items offset for current page

        Returns:
            Integer representing offset
        """
        return self.size * (self.page - 1)



from abc import abstractmethod
from typing import Any, Type, TypeVar
from abc import ABC

T = TypeVar("T")


class AbstractConverter(ABC):
    """
    Abstract class representing converter that is used to transform data from/to dictionary to/from Python classes
    """

    @abstractmethod
    def dump_to_dict(self, t: Any, remove_none: bool = False) -> dict[str, Any]:
        """
        Dumps class to dictionary

        Args:
            t: Object to convert
            remove_none: If True, after converting object to dictionary, all keys with None values will be removed

        Returns:
            Object converted to dictionary
        """
        pass

    @abstractmethod
    def load_from_dict(self, data: dict[str, Any] | list[dict[str, Any]], t: Type[T]) -> T:
        """
        Converts dictionary to object

        Args:
            data: Dict data
            t: Type to convert to

        Returns:
            Converted object
        """
        pass



import threading
import time
from logging import Logger

from orjson import orjson
from websocket import WebSocketApp

from lime_trader.converters.cattr_converter import converter
from lime_trader.models.accounts import AccountFeedAction, AccountFeedActionType


class AccountFeedClient(threading.Thread):
    """
    Client used to subscribe to account feeds.
    Should not be instantiated by end-user.
    """

    def __init__(self, websocket_app: WebSocketApp, logger: Logger):
        super().__init__()
        self._websocket_app = websocket_app
        self._logger = logger
        self._available = False

    def run(self):
        self._logger.info("Starting account feed thread")
        self._websocket_app.run_forever()

    def stop(self):
        self._logger.info("Stopping account feed thread")
        self._websocket_app.close()

    def on_account_streaming_feed_open(self, web_socket_app: WebSocketApp):
        self._logger.info("Account feed client connection opened")
        self._available = True

    def on_account_streaming_feed_close(self, web_socket_app: WebSocketApp):
        self._logger.info("Account feed client connection closed")
        self._available = False

    def _send_action(self, action: AccountFeedAction):
        self._logger.info(f"Account feed - sending action {action}")
        while not self._available or not self._websocket_app.sock:
            time.sleep(1)
        self._websocket_app.send(data=orjson.dumps(converter.dump_to_dict(action)))

    def subscribe_balance(self, account_number: str) -> None:
        """
        Subscribe to balance changes for specific account

        Args:
            account_number: Account number
        """
        action = AccountFeedAction(action=AccountFeedActionType.SUBSCRIBE_BALANCE, account=account_number)
        self._send_action(action=action)

    def subscribe_positions(self, account_number: str) -> None:
        """
        Subscribe to position changes for specific account

        Args:
            account_number: Account number
        """
        action = AccountFeedAction(action=AccountFeedActionType.SUBSCRIBE_POSITIONS, account=account_number)
        self._send_action(action=action)

    def subscribe_orders(self, account_number: str) -> None:
        """
        Subscribe to order changes for specific account

        Args:
            account_number: Account number
        """
        action = AccountFeedAction(action=AccountFeedActionType.SUBSCRIBE_ORDERS, account=account_number)
        self._send_action(action=action)

    def subscribe_trades(self, account_number: str) -> None:
        """
        Subscribe to trades changes for specific account

        Args:
            account_number: Account number
        """
        action = AccountFeedAction(action=AccountFeedActionType.SUBSCRIBE_TRADES, account=account_number)
        self._send_action(action=action)

    def unsubscribe_balance(self, account_number: str) -> None:
        """
        Unsubscribe from balance changes for specific account

        Args:
            account_number: Account number
        """
        action = AccountFeedAction(action=AccountFeedActionType.UNSUBSCRIBE_BALANCE, account=account_number)
        self._send_action(action=action)

    def unsubscribe_positions(self, account_number: str) -> None:
        """
        Unsubscribe from position changes for specific account

        Args:
            account_number: Account number
        """
        action = AccountFeedAction(action=AccountFeedActionType.UNSUBSCRIBE_POSITIONS, account=account_number)
        self._send_action(action=action)

    def unsubscribe_orders(self, account_number: str) -> None:
        """
        Unsubscribe from order changes for specific account

        Args:
            account_number: Account number
        """
        action = AccountFeedAction(action=AccountFeedActionType.UNSUBSCRIBE_ORDERS, account=account_number)
        self._send_action(action=action)

    def unsubscribe_trades(self, account_number: str) -> None:
        """
        Unsubscribe from trades changes for specific account

        Args:
            account_number: Account number
        """
        action = AccountFeedAction(action=AccountFeedActionType.UNSUBSCRIBE_TRADES, account=account_number)
        self._send_action(action=action)



from _decimal import Decimal
from logging import Logger

from lime_trader.api.authenticated_api_client import AuthenticatedApiClient
from lime_trader.exceptions.api_error import ApiError
from lime_trader.models.accounts import TradeSide
from lime_trader.models.errors import Error
from lime_trader.models.trading import (Order, PlaceOrderResponse, ValidateOrderResponse, CancelOrderResponse,
                                        OrderFee, OrderDetails)
from lime_trader.constants.urls import (TRADING_GET_ACTIVE_ORDERS, TRADING_PLACE_ORDER, TRADING_VALIDATE_ORDER,
                                        TRADING_GET_ORDER_DETAILS,
                                        TRADING_CANCEL_ORDER, TRADING_ESTIMATE_FEE_CHARGES)


class TradingClient:
    """
    Contains methods related to trading
    """

    def __init__(self, api_client: AuthenticatedApiClient, logger: Logger):
        """
        Contains methods related to trading

        Args:
            api_client: API client that will be used to execute all requests
            logger: Logger used to submit client log messages
        """
        self._api_client = api_client
        self._logger = logger

    def place_order(self, order: Order) -> PlaceOrderResponse:
        """
        Places an order. The order is accepted immediately, the method returns assigned id.
        The order is still validated with exactly same logic as in validate
        method and is sent to market on successful validation pass.
        Otherwise, the order will reject asynchronously, and you can query its status by calling the details' endpoint.

        Args:
            order: Order to place

        Returns:
            Order placing response with indicator of success and placed order id
        """
        return self._api_client.post(url=TRADING_PLACE_ORDER, json=order,
                                     path_params={},
                                     response_schema=PlaceOrderResponse)

    def validate_order(self, order: Order) -> ValidateOrderResponse:
        """
        Verifies an order and responds with the validation message if the order can not be placed at the moment.
        The order is not sent to market.

        Args:
            order: Order for verification

        Returns:
            Validation response
        """
        return self._api_client.post(url=TRADING_VALIDATE_ORDER, json=order,
                                     path_params={},
                                     response_schema=ValidateOrderResponse)

    def get_order_details(self, order_id: str) -> OrderDetails:
        """
        Get the order details by the specified order id

        Args:
            order_id: Order id

        Returns:
            Order details
        """
        return self._api_client.get(url=TRADING_GET_ORDER_DETAILS,
                                    path_params={"id": order_id}, params={},
                                    response_schema=OrderDetails)

    def cancel_order(self, order_id: str, message: str | None) -> CancelOrderResponse:
        """
        Cancels specified order

        Args:
            order_id: Order id
            message: Optional cancellation details

        Returns:
            Cancellation result with indicator of success
        """
        return self._api_client.post(url=TRADING_CANCEL_ORDER, path_params={"id": order_id}, json={"message": message},
                                     response_schema=CancelOrderResponse)

    def get_active_orders(self, account_number: str) -> list[OrderDetails]:
        """
        Get list of active orders for the account

        Args:
            account_number: Account number

        Returns:
            List of active orders
        """
        return self._api_client.get(url=TRADING_GET_ACTIVE_ORDERS, path_params={"account_number": account_number},
                                    params={},
                                    response_schema=list[OrderDetails])

    def estimate_fee_charges(self, account_number: str, symbol: str, quantity: Decimal, side: TradeSide,
                             price: Decimal) -> list[OrderFee]:
        """
        Gets estimated fees for specified order parameters, breaking down all charges by type

        Args:
            account_number: Account number
            symbol: Stock or an option
            quantity: Order quantity
            side: Order side
            price: Order price

        Returns:
            Estimated fees for specified order parameters by type
        """
        return self._api_client.post(url=TRADING_ESTIMATE_FEE_CHARGES, path_params={}, json={
            "account_number": account_number,
            "symbol": symbol,
            "quantity": quantity,
            "side": side,
            "price": price
        }, response_schema=list[OrderFee])

    def validate_and_place_order(self, order: Order) -> PlaceOrderResponse:
        validation_result = self.validate_order(order=order)
        if validation_result.is_valid:
            return self.place_order(order=order)
        raise ApiError(Error(code="validation", message=validation_result.validation_message,
                             status_code=None))



import datetime
from logging import Logger

from lime_trader.async_trader.api.async_authenticated_api_client import AsyncAuthenticatedApiClient
from lime_trader.clients.market_data_client import MarketDataClient
from lime_trader.models.market import (Quote, QuoteHistory, Period, Security, SecuritiesPage, Trade, TradesPage,
                                       CurrentSchedule)
from lime_trader.models.page import Page, PageRequest
from lime_trader.constants.urls import (MARKET_DATA_GET_CURRENT_QUOTE, MARKET_DATA_GET_QUOTES,
                                        MARKET_DATA_GET_TRADING_SCHEDULE,
                                        MARKET_DATA_LOOKUP_SECURITIES, MARKET_DATA_GET_TIME_AND_SALES,
                                        MARKET_DATA_GET_QUOTES_HISTORY)


class AsyncMarketDataClient(MarketDataClient):

    def __init__(self, api_client: AsyncAuthenticatedApiClient, logger: Logger):
        super().__init__(api_client=api_client, logger=logger)

    async def get_current_quote(self, symbol: str) -> Quote:
        return await self._api_client.get(MARKET_DATA_GET_CURRENT_QUOTE, path_params={},
                                          params={"symbol": symbol},
                                          response_schema=Quote)

    async def get_current_quotes(self, symbols: list[str]) -> list[Quote]:
        return await self._api_client.post(MARKET_DATA_GET_QUOTES, path_params={}, json=symbols,
                                           response_schema=list[Quote])

    async def get_quotes_history(self, symbol: str, period: Period, from_date: datetime.datetime,
                                 to_date: datetime.datetime) -> list[QuoteHistory]:
        return await self._api_client.get(MARKET_DATA_GET_QUOTES_HISTORY, path_params={},
                                          params={"symbol": symbol,
                                                  "period": period.value,
                                                  "from": from_date,
                                                  "to": to_date
                                                  }, response_schema=list[QuoteHistory])

    async def get_trading_schedule(self) -> CurrentSchedule:
        return await self._api_client.get(MARKET_DATA_GET_TRADING_SCHEDULE, path_params={}, params={},
                                          response_schema=CurrentSchedule)

    async def lookup_securities(self, query: str, page: PageRequest) -> Page[Security]:
        response = await self._api_client.get(MARKET_DATA_LOOKUP_SECURITIES, path_params={},
                                              params={"query": query,
                                                      "limit": page.size,
                                                      "skip": page.get_offset()},
                                              response_schema=SecuritiesPage)
        return Page(data=response.securities, number=page.page, size=page.size, total_elements=response.count)

    async def time_and_sales(self, symbol: str, from_date: datetime.datetime,
                             to_date: datetime.datetime, page: PageRequest) -> Page[Trade]:
        response = await self._api_client.get(MARKET_DATA_GET_TIME_AND_SALES, path_params={},
                                              params={"symbol": symbol,
                                                      "limit": page.size,
                                                      "skip": page.get_offset()},
                                              response_schema=TradesPage)
        return Page(data=response.trades, number=page.page, size=page.size, total_elements=response.count)



from urllib3.util import Url

HTTP_SCHEMA = "http"
WS_SCHEMA = "ws"

AUTHENTICATION_GET_TOKEN = Url(scheme=HTTP_SCHEMA, path="/connect/token")
AUTHENTICATION_AUTHORIZE = Url(scheme=HTTP_SCHEMA, path="/connect/authorize")

ACCOUNTS_GET_ACCOUNTS_BALANCES = Url(scheme=HTTP_SCHEMA, path="/accounts")
ACCOUNTS_STREAMING_FEED = Url(scheme=WS_SCHEMA, path="/accounts")
ACCOUNTS_GET_ACCOUNT_POSITIONS = Url(scheme=HTTP_SCHEMA, path="/accounts/{account_number}/positions")
ACCOUNTS_GET_ACCOUNT_TRADES = Url(scheme=HTTP_SCHEMA, path="/accounts/{account_number}/trades/{date}")
ACCOUNTS_GET_ROUTES = Url(scheme=HTTP_SCHEMA, path="/accounts/{account_number}/routes")
ACCOUNTS_GET_ACCOUNT_TRANSACTIONS = Url(scheme=HTTP_SCHEMA, path="/accounts/{account_number}/transactions")

TRADING_PLACE_ORDER = Url(scheme=HTTP_SCHEMA, path="/orders/place")
TRADING_VALIDATE_ORDER = Url(scheme=HTTP_SCHEMA, path="/orders/validate")
TRADING_GET_ORDER_DETAILS = Url(scheme=HTTP_SCHEMA, path="/orders/{id}")
TRADING_GET_ACTIVE_ORDERS = Url(scheme=HTTP_SCHEMA, path="/accounts/{account_number}/activeorders")
TRADING_CANCEL_ORDER = Url(scheme=HTTP_SCHEMA, path="/orders/{id}/cancel")
TRADING_ESTIMATE_FEE_CHARGES = Url(scheme=HTTP_SCHEMA, path="/pricing/fees")

MARKET_DATA_GET_CURRENT_QUOTE = Url(scheme=HTTP_SCHEMA, path="/marketdata/quote")
MARKET_DATA_GET_QUOTES = Url(scheme=HTTP_SCHEMA, path="/marketdata/quotes")
MARKET_DATA_GET_QUOTES_HISTORY = Url(scheme=HTTP_SCHEMA, path="/marketdata/history")
MARKET_DATA_GET_TRADING_SCHEDULE = Url(scheme=HTTP_SCHEMA, path="/marketdata/schedule")
MARKET_DATA_LOOKUP_SECURITIES = Url(scheme=HTTP_SCHEMA, path="/securities")
MARKET_DATA_GET_TIME_AND_SALES = Url(scheme=HTTP_SCHEMA, path="/marketdata/trades")



import logging

from lime_trader.async_trader.clients.async_account_client import AsyncAccountClient
from lime_trader.async_trader.clients.async_market_data_client import AsyncMarketDataClient
from lime_trader.async_trader.clients.async_trading_client import AsyncTradingClient
from lime_trader.constants.config import DEFAULT_TIMEOUT
from lime_trader.converters.cattr_converter import CAttrConverter

from lime_trader import LimeClient
from lime_trader.async_trader.api.async_api_client import AsyncApiClient
from lime_trader.async_trader.api.async_authenticated_api_client import AsyncAuthenticatedApiClient
from lime_trader.models.accounts import Credentials
from lime_trader.models.token_storage import AsyncApiTokenStorage


class AsyncLimeClient(LimeClient):
    def __init__(self, base_url: str, credentials: Credentials, logger: logging.Logger, auth_url: str = None,
                 timeout: int = DEFAULT_TIMEOUT):
        super().__init__(base_url=base_url, credentials=credentials, logger=logger, auth_url=auth_url)
        self._converter = CAttrConverter()
        self._api_client = AsyncApiClient(base_url=self._base_url, converter=self._converter,
                                          use_https=self._api_use_https,
                                          use_wss=self._api_use_wss, logger=self._logger, timeout=timeout)
        self._auth_api_client = AsyncApiClient(base_url=self._auth_url, converter=self._converter,
                                               use_https=self._auth_use_https,
                                               use_wss=self._auth_use_wss, logger=self._logger, timeout=timeout)
        self._token_storage = AsyncApiTokenStorage(credentials=credentials, api_client=self._auth_api_client)
        self._authenticated_api_client = AsyncAuthenticatedApiClient(api_client=self._api_client,
                                                                     token_storage=self._token_storage,
                                                                     credentials=self._credentials, logger=self._logger)

        # instantiate clients
        self.account = AsyncAccountClient(api_client=self._authenticated_api_client, logger=self._logger)
        self.trading = AsyncTradingClient(api_client=self._authenticated_api_client, logger=self._logger)
        self.market = AsyncMarketDataClient(api_client=self._authenticated_api_client, logger=self._logger)
